# encoding: utf-8

"""Integrations tests for headings and subtotals behaviors."""

import numpy as np

from cr.cube.crunch_cube import CrunchCube

from unittest import TestCase

from ...fixtures import CR


class TestHeadersAndSubtotals(TestCase):
    def test_headings_econ_blame_one_subtotal(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = [
            "President Obama",
            "Republicans in Congress",
            "Test New Heading (Obama and Republicans)",
            "Both",
            "Neither",
            "Not sure",
        ]
        actual = cube.labels(include_transforms_for_dims=[0])[0]
        self.assertEqual(actual, expected)

    def test_headings_econ_blame_one_subtotal_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = [
            "President Obama",
            "Republicans in Congress",
            "Both",
            "Neither",
            "Not sure",
        ]
        actual = cube.labels(include_transforms_for_dims=None)[0]
        self.assertEqual(actual, expected)

    def test_headings_econ_blame_two_subtotal_without_missing(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING)
        expected = [
            "President Obama",
            "Republicans in Congress",
            "Test New Heading (Obama and Republicans)",
            "Both",
            "Neither",
            "Not sure",
            "Test Heading with Skipped",
        ]
        actual = cube.labels(include_transforms_for_dims=[0])[0]
        self.assertEqual(actual, expected)

    def test_headings_two_subtotal_without_missing_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING)
        expected = [
            "President Obama",
            "Republicans in Congress",
            "Both",
            "Neither",
            "Not sure",
        ]
        actual = cube.labels(include_transforms_for_dims=None)[0]
        self.assertEqual(actual, expected)

    def test_headings_econ_blame_two_subtotal_with_missing(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING)
        expected = [
            "President Obama",
            "Republicans in Congress",
            "Test New Heading (Obama and Republicans)",
            "Both",
            "Neither",
            "Not sure",
            "Skipped",
            "Not Asked",
            "No Data",
            "Test Heading with Skipped",
        ]
        actual = cube.labels(include_missing=True, include_transforms_for_dims=[0])[0]
        self.assertEqual(actual, expected)

    def test_subtotals_as_array_one_transform(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = np.array([285, 396, 681, 242, 6, 68])
        actual = cube.as_array(include_transforms_for_dims=[0])
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_as_array_one_transform_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = np.array([285, 396, 242, 6, 68])
        actual = cube.as_array(include_transforms_for_dims=None)
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_as_array_two_transforms_missing_excluded(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING)
        expected = np.array([285, 396, 681, 242, 6, 68, 74])
        actual = cube.as_array(include_transforms_for_dims=[0])
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_proportions_one_transform(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = np.array(
            [0.2858576, 0.3971916, 0.6830491, 0.2427282, 0.0060181, 0.0682046]
        )
        actual = cube.proportions(include_transforms_for_dims=[0])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_one_transform_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = np.array([0.2858576, 0.3971916, 0.2427282, 0.0060181, 0.0682046])
        actual = cube.proportions(include_transforms_for_dims=None)
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_two_transforms_missing_excluded(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS_MISSING)
        expected = np.array(
            [
                0.2858576,
                0.3971916,
                0.6830491,
                0.2427282,
                0.0060181,
                0.0682046,
                0.0742227,
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0])
        np.testing.assert_almost_equal(actual, expected)

    def test_labels_on_2d_cube_with_hs_on_1st_dim(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = [
            [
                "President Obama",
                "Republicans in Congress",
                "Test New Heading (Obama and Republicans)",
                "Both",
                "Neither",
                "Not sure",
            ],
            [
                "Very liberal",
                "Liberal",
                "Moderate",
                "Conservative",
                "Very Conservative",
                "Not sure",
            ],
        ]
        actual = cube.labels(include_transforms_for_dims=[0, 1])
        self.assertEqual(actual, expected)

    def test_labels_on_2d_cube_with_hs_on_both_dim(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = [
            [
                "President Obama",
                "Republicans in Congress",
                "Test New Heading (Obama and Republicans)",
                "Both",
                "Neither",
                "Not sure",
            ],
            [
                "Very liberal",
                "Liberal",
                "Moderate",
                "Test 2nd dim Heading",
                "Conservative",
                "Very Conservative",
                "Not sure",
            ],
        ]
        actual = cube.labels(include_transforms_for_dims=[0, 1])
        self.assertEqual(actual, expected)

    def test_labels_on_2d_cube_with_hs_on_both_dim_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = [
            [
                "President Obama",
                "Republicans in Congress",
                "Both",
                "Neither",
                "Not sure",
            ],
            [
                "Very liberal",
                "Liberal",
                "Moderate",
                "Conservative",
                "Very Conservative",
                "Not sure",
            ],
        ]
        actual = cube.labels(include_transforms_for_dims=None)
        self.assertEqual(actual, expected)

    def test_subtotals_as_array_2d_cube_with_hs_on_row(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array(
            [
                [3, 14, 80, 114, 67, 7],
                [59, 132, 162, 29, 12, 2],
                [62, 146, 242, 143, 79, 9],
                [6, 29, 109, 67, 26, 5],
                [1, 1, 1, 1, 0, 2],
                [3, 6, 23, 7, 5, 24],
            ]
        )
        actual = cube.as_array(include_transforms_for_dims=[0, 1])
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_as_array_2d_cube_with_hs_on_col(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_COL_HS)
        expected = np.array(
            [
                [3, 14, 80, 94, 114, 67, 7],
                [59, 132, 162, 294, 29, 12, 2],
                [6, 29, 109, 138, 67, 26, 5],
                [1, 1, 1, 2, 1, 0, 2],
                [3, 6, 23, 29, 7, 5, 24],
            ]
        )
        actual = cube.as_array(include_transforms_for_dims=[0, 1])
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_as_array_2d_cube_with_hs_on_both_dim(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = np.array(
            [
                [3, 14, 80, 94, 114, 67, 7],
                [59, 132, 162, 294, 29, 12, 2],
                [62, 146, 242, 388, 143, 79, 9],
                [6, 29, 109, 138, 67, 26, 5],
                [1, 1, 1, 2, 1, 0, 2],
                [3, 6, 23, 29, 7, 5, 24],
            ]
        )
        actual = cube.as_array(include_transforms_for_dims=[0, 1])
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_as_array_2d_cube_with_hs_on_both_dim_do_not_fetch(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = np.array(
            [
                [3, 14, 80, 114, 67, 7],
                [59, 132, 162, 29, 12, 2],
                [6, 29, 109, 67, 26, 5],
                [1, 1, 1, 1, 0, 2],
                [3, 6, 23, 7, 5, 24],
            ]
        )
        actual = cube.as_array(include_transforms_for_dims=None)
        np.testing.assert_array_equal(actual, expected)

    def test_subtotals_margin_2d_cube_with_hs_on_row_by_col(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array([72, 182, 375, 218, 110, 40])
        actual = cube.margin(axis=0, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_margin_2d_cube_with_hs_on_row_by_row(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array([285, 396, 681, 242, 6, 68])
        actual = cube.margin(axis=1, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_margin_2d_cube_with_hs_on_two_dim_by_col(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = np.array([72, 182, 375, 557, 218, 110, 40])
        actual = cube.margin(axis=0, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_margin_2d_cube_with_hs_on_two_dim_by_row(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = np.array([285, 396, 681, 242, 6, 68])
        actual = cube.margin(axis=1, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_2d_cube_with_hs_on_row_by_cell(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array(
            [
                [0.00300903, 0.01404213, 0.08024072, 0.11434303, 0.0672016, 0.00702106],
                [
                    0.05917753,
                    0.13239719,
                    0.16248746,
                    0.02908726,
                    0.01203611,
                    0.00200602,
                ],
                [
                    0.06218656,
                    0.14643932,
                    0.24272818,
                    0.14343029,
                    0.07923771,
                    0.00902708,
                ],
                [0.00601805, 0.02908726, 0.10932798, 0.0672016, 0.02607823, 0.00501505],
                [0.00100301, 0.00100301, 0.00100301, 0.00100301, 0, 0.00200602],
                [
                    0.00300903,
                    0.00601805,
                    0.02306921,
                    0.00702106,
                    0.00501505,
                    0.02407222,
                ],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_2d_cube_with_hs_on_row_by_col(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array(
            [
                [0.04166667, 0.07692308, 0.21333333, 0.52293578, 0.60909091, 0.175],
                [0.81944444, 0.72527473, 0.432, 0.13302752, 0.10909091, 0.05],
                [0.86111111, 0.8021978, 0.64533333, 0.6559633, 0.71818182, 0.225],
                [0.08333333, 0.15934066, 0.29066667, 0.30733945, 0.23636364, 0.125],
                [0.01388889, 0.00549451, 0.00266667, 0.00458716, 0, 0.05],
                [0.04166667, 0.03296703, 0.06133333, 0.03211009, 0.04545455, 0.6],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1], axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_2d_cube_with_hs_on_row_by_row(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array(
            [
                [0.01052632, 0.04912281, 0.28070175, 0.4, 0.23508772, 0.0245614],
                [0.1489899, 0.33333333, 0.40909091, 0.07323232, 0.03030303, 0.00505051],
                [0.09104258, 0.2143906, 0.35535977, 0.20998532, 0.11600587, 0.01321586],
                [0.02479339, 0.11983471, 0.45041322, 0.2768595, 0.10743802, 0.02066116],
                [0.16666667, 0.16666667, 0.16666667, 0.16666667, 0, 0.33333333],
                [
                    0.04411765,
                    0.08823529,
                    0.33823529,
                    0.10294118,
                    0.07352941,
                    0.35294118,
                ],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1], axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_proportions_2d_cube_with_hs_on_two_dim_by_cell(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_AND_COL_HS)
        expected = np.array(
            [
                [
                    0.00300903,
                    0.01404213,
                    0.08024072,
                    0.09428285,
                    0.11434303,
                    0.0672016,
                    0.00702106,
                ],
                [
                    0.05917753,
                    0.13239719,
                    0.16248746,
                    0.29488465,
                    0.02908726,
                    0.01203611,
                    0.00200602,
                ],
                [
                    0.06218656,
                    0.14643932,
                    0.24272818,
                    0.3891675,
                    0.14343029,
                    0.07923771,
                    0.00902708,
                ],
                [
                    0.00601805,
                    0.02908726,
                    0.10932798,
                    0.13841525,
                    0.0672016,
                    0.02607823,
                    0.00501505,
                ],
                [
                    0.00100301,
                    0.00100301,
                    0.00100301,
                    0.00200602,
                    0.00100301,
                    0,
                    0.00200602,
                ],
                [
                    0.00300903,
                    0.00601805,
                    0.02306921,
                    0.02908726,
                    0.00702106,
                    0.00501505,
                    0.02407222,
                ],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_labels_with_hs(self):
        cube = CrunchCube(CR.SIMPLE_CA_HS)
        expected = [
            ["ca_subvar_1", "ca_subvar_2", "ca_subvar_3"],
            ["a", "b", "Test A and B combined", "c", "d"],
        ]
        actual = cube.labels(include_transforms_for_dims=[0])
        self.assertEqual(actual, expected)

    def test_ca_as_array_with_hs(self):
        cube = CrunchCube(CR.SIMPLE_CA_HS)
        expected = np.array([[3, 3, 6, 0, 0], [1, 3, 4, 2, 0], [0, 2, 2, 1, 3]])
        actual = cube.as_array(include_transforms_for_dims=[0, 1])
        np.testing.assert_array_equal(actual, expected)

    def test_ca_proportions_with_hs(self):
        cube = CrunchCube(CR.SIMPLE_CA_HS)
        expected = np.array(
            [
                [0.5, 0.5, 1, 0, 0],
                [0.16666667, 0.5, 0.66666667, 0.33333333, 0],
                [0, 0.33333333, 0.33333333, 0.16666667, 0.5],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1], axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_margin_with_hs(self):
        cube = CrunchCube(CR.SIMPLE_CA_HS)
        expected = np.array([6, 6, 6])
        actual = cube.margin(include_transforms_for_dims=[0, 1], axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_hs_with_anchor_on_zero_position_labels(self):
        cube = CrunchCube(CR.ECON_US_PROBLEM_X_BIGGER_PROBLEM)
        expected = [
            [
                "Serious net",
                "Very serious",
                "Somewhat serious",
                "Not very serious",
                "Not at all serious",
                "Not sure",
            ],
            [
                "Sexual assaults that go unreported or unpunished",
                "False accusations of sexual assault",
            ],
        ]
        actual = cube.labels(include_transforms_for_dims=[0, 1])
        self.assertEqual(actual, expected)

    def test_hs_with_anchor_on_zero_position_as_props_by_col(self):
        cube = CrunchCube(CR.ECON_US_PROBLEM_X_BIGGER_PROBLEM)
        expected = np.array(
            [
                [0.93244626, 0.66023166],
                [0.63664278, 0.23166023],
                [0.29580348, 0.42857143],
                [0.04401228, 0.21428571],
                [0.00307062, 0.06177606],
                [0.02047083, 0.06370656],
            ]
        )
        actual = cube.proportions(axis=0, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_hs_with_anchor_on_zero_position_as_props_by_row(self):
        cube = CrunchCube(CR.ECON_US_PROBLEM_X_BIGGER_PROBLEM)
        expected = np.array(
            [
                [0.72705507, 0.27294493],
                [0.83827493, 0.16172507],
                [0.56555773, 0.43444227],
                [0.27922078, 0.72077922],
                [0.08571429, 0.91428571],
                [0.37735849, 0.62264151],
            ]
        )
        actual = cube.proportions(axis=1, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_hs_with_anchor_on_zero_position_as_props_by_cell(self):
        cube = CrunchCube(CR.ECON_US_PROBLEM_X_BIGGER_PROBLEM)
        expected = np.array(
            [
                [0.60936455, 0.22876254],
                [0.41605351, 0.08026756],
                [0.19331104, 0.14849498],
                [0.02876254, 0.07424749],
                [0.00200669, 0.02140468],
                [0.01337793, 0.02207358],
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_subtotals_pvals_2d_cube_with_hs_on_row(self):
        """Ensure that pvals shape is the same as table shape with H%S"""
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = np.array(
            [
                [
                    1.92562832e-06,
                    5.20117283e-12,
                    8.30737469e-05,
                    0.00000000e00,
                    1.77635684e-15,
                    1.13223165e-01,
                ],
                [
                    2.90878432e-14,
                    0.00000000e00,
                    8.11477145e-02,
                    0.00000000e00,
                    5.87376814e-11,
                    4.64141147e-06,
                ],
                [
                    1.05605732e-03,
                    3.70613426e-03,
                    6.11851617e-03,
                    1.18269053e-02,
                    8.68890220e-01,
                    7.62914197e-02,
                ],
                [
                    3.69990005e-01,
                    9.19546240e-01,
                    2.88068221e-01,
                    7.57299844e-01,
                    3.86924216e-01,
                    2.41648361e-04,
                ],
                [
                    3.53745446e-01,
                    3.70094812e-02,
                    5.03974440e-01,
                    1.67769523e-02,
                    3.15641644e-01,
                    0.00000000e00,
                ],
            ]
        )
        actual = cube.pvals()
        np.testing.assert_almost_equal(actual, expected)

    def test_fruit_hs_top_bottom_labels(self):
        cube = CrunchCube(CR.FRUIT_HS_TOP_BOTTOM)
        expected = [["TOP", "rambutan", "MIDDLE", "satsuma", "BOTTOM"]]
        actual = cube.labels(include_transforms_for_dims=[0])
        assert actual == expected

    def test_fruit_hs_top_bottom_inserted_indices(self):
        cube = CrunchCube(CR.FRUIT_HS_TOP_BOTTOM)
        expected = [[0, 2, 4]]
        actual = cube.inserted_hs_indices(prune=True)
        assert actual == expected

    def test_fruit_hs_top_bottom_counts(self):
        cube = CrunchCube(CR.FRUIT_HS_TOP_BOTTOM)
        expected = np.array([100, 33, 100, 67, 100])
        actual = cube.as_array(include_transforms_for_dims=[0])
        np.testing.assert_array_equal(actual, expected)

    def test_fruit_x_pets_hs_top_bottom_middle_props(self):
        cube = CrunchCube(CR.FRUIT_X_PETS_HS_TOP_BOTTOM)
        expected = np.array(
            [
                [1.0, 1.0, 1.0],
                [0.3, 0.35294118, 0.31578947],
                [1.0, 1.0, 1.0],
                [0.7, 0.64705882, 0.68421053],
                [1.0, 1.0, 1.0],
            ]
        )
        actual = cube.proportions(axis=0, include_transforms_for_dims=[0])
        np.testing.assert_almost_equal(actual, expected)

    def test_fruit_x_pets_hs_top_bottom_middle_counts(self):
        cube = CrunchCube(CR.FRUIT_X_PETS_HS_TOP_BOTTOM)
        expected = np.array(
            [[40, 34, 38], [12, 12, 12], [40, 34, 38], [28, 22, 26], [40, 34, 38]]
        )
        actual = cube.as_array(include_transforms_for_dims=[0])
        np.testing.assert_array_equal(actual, expected)

    def test_hs_indices_pruned_cat_x_date(self):
        cube = CrunchCube(CR.CAT_X_DATE_HS_PRUNE)
        expected = [0, 3, 6]
        actual = cube.inserted_hs_indices(prune=True)[0]
        assert actual == expected

    def test_hs_indices_pruned_cat_x_num(self):
        cube = CrunchCube(CR.CAT_X_NUM_HS_PRUNE)
        expected = [0, 1, 3]
        actual = cube.inserted_hs_indices(prune=True)[0]
        assert actual == expected

        # Ensure cached properties are not updated
        actual = cube.inserted_hs_indices(prune=True)[0]
        assert actual == expected
        actual = cube.inserted_hs_indices(prune=True)[0]
        assert actual == expected

    def test_cat_x_num_counts_pruned_with_hs(self):
        cube = CrunchCube(CR.CAT_X_NUM_HS_PRUNE)
        expected = np.array([[0], [1], [1], [0]])
        # Extract only non-masked (pruned) values
        table = cube.as_array(include_transforms_for_dims=[0], prune=True)
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_array_equal(actual, expected)

    def test_cat_x_num_counts_pruned_without_hs(self):
        cube = CrunchCube(CR.CAT_X_NUM_HS_PRUNE)
        expected = np.array([[1]])
        table = cube.as_array(prune=True)
        # Extract only non-masked (pruned) values
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_array_equal(actual, expected)

    def test_mr_x_cat_hs_counts(self):
        cube = CrunchCube(CR.PETS_X_FRUIT_HS)
        expected = np.array([[12, 28, 40], [12, 22, 34], [12, 26, 38]])
        actual = cube.as_array(include_transforms_for_dims=[0, 1])
        np.testing.assert_array_equal(actual, expected)

    def test_mr_x_cat_hs_props_by_cell(self):
        cube = CrunchCube(CR.PETS_X_FRUIT_HS)
        # TODO: Change expectation once the MR cell props are fixed.
        expected = (3, 3)
        actual = cube.proportions(axis=None, include_transforms_for_dims=[0, 1]).shape
        np.testing.assert_array_equal(actual, expected)

    def test_mr_x_cat_hs_props_by_row(self):
        cube = CrunchCube(CR.PETS_X_FRUIT_HS)
        # TODO: Change expectation once the MR cell props are fixed.
        expected = (3, 3)
        actual = cube.proportions(axis=0, include_transforms_for_dims=[0, 1]).shape
        np.testing.assert_array_equal(actual, expected)

    def test_mr_x_cat_hs_props_by_col(self):
        cube = CrunchCube(CR.PETS_X_FRUIT_HS)
        # TODO: Change expectation once the MR cell props are fixed.
        expected = (3, 3)
        actual = cube.proportions(axis=1, include_transforms_for_dims=[0, 1]).shape
        np.testing.assert_array_equal(actual, expected)

    def test_missing_cat_hs_labels(self):
        cube = CrunchCube(CR.MISSING_CAT_HS)

        # Don't expect the missing category "Non voters"
        expected = [
            [
                "Whites",
                "White college women voters",
                "White non-college women voters",
                "White college men voters",
                "White non-college men voters",
                "Black voters",
                "Latino and other voters",
            ]
        ]
        actual = cube.labels(include_transforms_for_dims=[0])
        assert actual == expected

    def test_ca_x_cat_counts_with_hs(self):
        cube = CrunchCube(CR.CA_X_CAT_HS)

        # Assert counts without H&S
        expected = np.array(
            [
                [[1, 1, 0, 0, 0], [0, 0, 1, 1, 1], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]],
                [[1, 0, 0, 0, 0], [0, 1, 0, 1, 0], [0, 0, 1, 0, 1], [0, 0, 0, 0, 0]],
                [[0, 0, 0, 0, 0], [1, 0, 0, 1, 0], [0, 1, 0, 0, 0], [0, 0, 1, 0, 1]],
            ]
        )
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

        # Assert counts with H&S
        expected = np.array(
            [
                [
                    [1, 1, 0, 2, 0, 0, 0],
                    [0, 0, 1, 1, 1, 1, 2],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                [
                    [1, 0, 0, 1, 0, 0, 0],
                    [0, 1, 0, 1, 1, 0, 1],
                    [0, 0, 1, 1, 0, 1, 1],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 1, 1, 0, 1],
                    [0, 1, 0, 1, 0, 0, 0],
                    [0, 0, 1, 1, 0, 1, 1],
                ],
            ]
        )

        # Include transforms for all CA and CAT dims (hence 0, 1 and 2)
        actual = cube.as_array(include_transforms_for_dims=[0, 1, 2])
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_cat_margin_with_hs(self):
        cube = CrunchCube(CR.CA_X_CAT_HS)

        # Assert counts without H&S
        expected = np.array([[1, 1, 1, 1, 1], [1, 1, 1, 1, 1], [1, 1, 1, 1, 1]])
        actual = cube.margin(axis=1, include_transforms_for_dims=[1])
        np.testing.assert_array_equal(actual, expected)

    def test_cat_x_items_x_cats_margin_with_hs(self):
        cube = CrunchCube(CR.CAT_X_ITEMS_X_CATS_HS)

        # Assert counts without H&S
        expected = np.array(
            [
                [
                    1287.9364594075469,
                    2050.0571926339885,
                    782.9403891997617,
                    225.4066607421201,
                    2622.8036855384603,
                    974.5889537143403,
                    490.5036709315041,
                    373.8221357520375,
                ],
                [
                    1147.3697583254452,
                    2557.8859179678857,
                    1096.7841912034742,
                    374.0411471364339,
                    1876.3400274431515,
                    1002.2399030962134,
                    457.92228898229905,
                    419.5110527202654,
                ],
                [
                    1053.855075581148,
                    2699.612841209989,
                    1427.7399174151794,
                    380.8205091587366,
                    1027.7782011616534,
                    606.7100283028576,
                    218.42735718966821,
                    265.29362712412535,
                ],
            ]
        )
        actual = cube.margin(axis=2, include_transforms_for_dims=[1, 2])
        np.testing.assert_almost_equal(actual, expected)

    def test_cat_x_mr_weighted_with_hs(self):
        cube = CrunchCube(CR.CAT_X_MR_WEIGHTED_HS)
        expected = np.array(
            [
                [
                    [0.05865163, 0.087823, 0.07486857, 0.0735683, 0.0, 0.08148267],
                    [
                        0.20246563,
                        0.33500382,
                        0.33176765,
                        0.27870974,
                        0.36309359,
                        0.33341993,
                    ],
                    [
                        0.54592009,
                        0.45988528,
                        0.49802406,
                        0.48137697,
                        0.51250032,
                        0.47855168,
                    ],
                    [
                        0.1051054,
                        0.06727875,
                        0.0733213,
                        0.12392602,
                        0.12440609,
                        0.07023618,
                    ],
                    [0.05508246, 0.04566041, 0.01379632, 0.02729368, 0.0, 0.03006505],
                    [0.03277479, 0.00434874, 0.0082221, 0.01512529, 0.0, 0.00624449],
                    [
                        0.26111726,
                        0.42282682,
                        0.40663622,
                        0.35227804,
                        0.36309359,
                        0.4149026,
                    ],
                    [
                        0.80703735,
                        0.8827121,
                        0.90466028,
                        0.83365501,
                        0.87559391,
                        0.89345428,
                    ],
                ],
                [
                    [0.00235883, 0.01361576, 0.01328221, 0.01212187, 0.0, 0.01345251],
                    [0.13002321, 0.0591588, 0.09857174, 0.05056353, 0.0, 0.07844882],
                    [0.65429951, 0.77915194, 0.74437239, 0.61537442, 1.0, 0.76212966],
                    [0.13730378, 0.11171429, 0.11961331, 0.26739934, 0.0, 0.11558033],
                    [0.04323988, 0.02601641, 0.01593825, 0.02729368, 0.0, 0.02108382],
                    [0.03277479, 0.0103428, 0.0082221, 0.02724716, 0.0, 0.00930486],
                    [0.13238204, 0.07277456, 0.11185395, 0.06268541, 0.0, 0.09190133],
                    [0.78668155, 0.85192649, 0.85622634, 0.67805982, 1.0, 0.85403098],
                ],
                [
                    [0.003676, 0.00486795, 0.0082221, 0.01212187, 0.0, 0.00650959],
                    [
                        0.03884185,
                        0.00625322,
                        0.02928964,
                        0.02729368,
                        0.12440609,
                        0.01752802,
                    ],
                    [
                        0.39625335,
                        0.4717925,
                        0.44918748,
                        0.46124151,
                        0.40543568,
                        0.46072884,
                    ],
                    [
                        0.4255265,
                        0.44274565,
                        0.42191512,
                        0.43268073,
                        0.47015822,
                        0.43255049,
                    ],
                    [0.1054366, 0.07434068, 0.08316356, 0.05153692, 0.0, 0.07865889],
                    [0.03026569, 0.0, 0.0082221, 0.01512529, 0.0, 0.00402417],
                    [
                        0.04251785,
                        0.01112118,
                        0.03751174,
                        0.03941555,
                        0.12440609,
                        0.02403761,
                    ],
                    [
                        0.4387712,
                        0.48291368,
                        0.48669922,
                        0.50065705,
                        0.52984178,
                        0.48476645,
                    ],
                ],
            ]
        )
        actual = cube.proportions(axis=1, include_transforms_for_dims=[0, 1, 2])
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_x_ca_props_by_row_without_hs(self):
        cube = CrunchCube(CR.MR_X_CA_HS)
        expected = np.array(
            [
                [
                    [0.66666667, 0.33333333, 0.00000000, 0.00000000],
                    [0.33333333, 0.33333333, 0.33333333, 0.00000000],
                    [0.00000000, 0.33333333, 0.33333333, 0.33333333],
                ],
                [
                    [0.50000000, 0.50000000, 0.00000000, 0.00000000],
                    [0.25000000, 0.25000000, 0.50000000, 0.00000000],
                    [0.00000000, 0.25000000, 0.00000000, 0.75000000],
                ],
                [
                    [np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan],
                ],
            ]
        )
        with self.assertRaises(ValueError):
            # "Table" direction not allowed, because cube's rows are CA dim
            actual = cube.proportions()
        actual = cube.proportions(axis=2)
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_x_ca_props_by_row_with_hs(self):
        cube = CrunchCube(CR.MR_X_CA_HS)
        expected = np.array(
            [
                [
                    [
                        0.66666667,
                        0.33333333,
                        1.00000000,
                        0.00000000,
                        0.00000000,
                        0.00000000,
                        1.00000000,
                    ],
                    [
                        0.33333333,
                        0.33333333,
                        0.66666667,
                        0.33333333,
                        0.00000000,
                        0.33333333,
                        1.00000000,
                    ],
                    [
                        0.00000000,
                        0.33333333,
                        0.33333333,
                        0.33333333,
                        0.33333333,
                        0.66666667,
                        1.00000000,
                    ],
                ],
                [
                    [
                        0.50000000,
                        0.50000000,
                        1.00000000,
                        0.00000000,
                        0.00000000,
                        0.00000000,
                        1.00000000,
                    ],
                    [
                        0.25000000,
                        0.25000000,
                        0.50000000,
                        0.50000000,
                        0.00000000,
                        0.50000000,
                        1.00000000,
                    ],
                    [
                        0.00000000,
                        0.25000000,
                        0.25000000,
                        0.00000000,
                        0.75000000,
                        0.75000000,
                        1.00000000,
                    ],
                ],
                [
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                ],
            ]
        )

        with self.assertRaises(ValueError):
            # "Table" direction not allowed, because cube's rows are CA dim
            actual = cube.proportions(include_transforms_for_dims=[0, 1, 2])

        actual = cube.proportions(axis=2, include_transforms_for_dims=[0, 1, 2])
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_cat_x_mr_x_ca_subvar_counts_pruning(self):
        cube = CrunchCube(CR.CA_CAT_X_MR_X_CA_SUBVAR_HS)
        expected = np.array(
            [
                [False, False, True],
                [False, False, True],
                [False, False, True],
                [True, True, True],
            ]
        )
        actual = cube.as_array(prune=True)[0].mask
        np.testing.assert_array_equal(actual, expected)

    def test_ca_cat_x_mr_x_ca_subvar_proportions_pruning(self):
        cube = CrunchCube(CR.CA_CAT_X_MR_X_CA_SUBVAR_HS)
        expected = np.array(
            [
                [False, False, True],
                [False, False, True],
                [False, False, True],
                [True, True, True],
            ]
        )

        with self.assertRaises(ValueError):
            # "Table" direction not allowed cuz CA items
            actual = cube.proportions(prune=True)[0].mask

        actual = cube.proportions(axis=1, prune=True)[0].mask
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_mr_counts_pruning(self):
        cube = CrunchCube(CR.CA_X_MR_HS)
        actual = cube.as_array(prune=True)[0].mask
        expected = np.array(
            [
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
            ]
        )
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_mr_proportions_pruning(self):
        cube = CrunchCube(CR.CA_X_MR_HS)
        expected = np.array(
            [
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
            ]
        )
        actual = cube.proportions(axis=None, prune=True)[0].mask
        np.testing.assert_array_equal(actual, expected)



# encoding: utf-8

import warnings

import numpy as np

from ..util import lazyproperty


class SingleSidedMovingAvgSmoother(object):
    """Create and configure smoothing function for one-sided moving average."""

    def __init__(self, partition, measure_expr):
        self._partition = partition
        self._measure_expression = measure_expr

    @lazyproperty
    def values(self):
        """1D/2D float64 ndarray of smoothed values including additional nans.

        Given a series of numbers and a fixed subset size, the first element of the
        moving average is obtained by taking the average of the initial fixed subset
        of the number series. Then the subset is modified by `shifting forward` the
        values. A moving average is commonly used with time series data to smooth
        out short-term fluctuations and highlight longer-term trends or cycles.

        The below examples will show 1D and 2D array rolling mean calculations with
        window sizes of 2 and 3, respectively.

                                    [window = 2]
        ----------------------------------------------------------------------------
            x    |   smooth(x)                  x     |        smooth(x)
        ---------+--------------         -------------+------------------------
            1    |    NaN                 1  3  2  3  |   NaN  2.0  2.5  2.5
            2    |    1.5                 2  3  3  2  |   NaN  2.5  3.0  2.5
            3    |    2.5                 3  2  4  4  |   NaN  2.5  3.0  4.0
            4    |    3.5                 4  1  5  1  |   NaN  2.5  3.0  3.0

                                    [window = 3]
        ----------------------------------------------------------------------------
            x    |   smooth(x)                  x     |        smooth(x)
        ---------+--------------         -------------+------------------------
            1    |    NaN                 1  3  2  3  |   NaN  NaN   2.0  2.67
            2    |    NaN                 2  3  3  2  |   NaN  NaN  2.67  2.67
            3    |     2                  3  2  4  4  |   NaN  NaN   3.0  3.33
            4    |     3                  4  1  5  1  |   NaN  NaN  3.33  2.33

        This is performed just taking the average of the last 2 or 3 rows according
        to the window, all the way down the column.
        """
        values = self._base_values
        if not self._can_smooth:
            # return original values if smoothing cannot be performed
            return values
        smoothed_values = self._smoothed_values
        # offset between original values and smoothed values
        offset = [values.shape[-1] - smoothed_values.shape[-1]]
        additional_nans = np.full(list(values.shape[:-1]) + offset, np.nan)
        return np.concatenate([additional_nans, smoothed_values], axis=values.ndim - 1)

    @lazyproperty
    def _base_measure(self):
        """str base_measure parameter specified in the measure expression"""
        base_measure = self._measure_expression.get("base_measure")
        # --- Mapping of measure-keywords to `partition` property-names ---
        measure_member = {
            "col_percent": "column_percentages",
            "col_index": "column_index",
            "mean": "means",
            "scale_mean": "scale_means_row",
        }
        if base_measure not in measure_member.keys():
            raise ValueError(
                "Base measure not recognized. Allowed values: 'col_percent', "
                "'col_index', 'mean', 'scale_mean', got: {}.".format(base_measure)
            )
        return measure_member.get(base_measure)

    @lazyproperty
    def _base_values(self):
        """ndarray base-measure values from the partition.

        The `base_measure` is expressed in the kwargs of the function_spec and used
        to get the values for the partition.
        """
        return np.array(getattr(self._partition, self._base_measure))

    @lazyproperty
    def _can_smooth(self):
        """bool, true if current data is smoothable.

        If the measure_expression parameters describe valid smoothing and the column
        dimensions is a categorical date it returns True.
        """
        # --- window cannot be wider than the number of periods in data
        # --- and it must be at least 2.
        if self._window > self._base_values.shape[-1] or self._window < 2:
            warnings.warn(
                "No smoothing performed. Smoothing window must be between 2 and the "
                "number of periods ({}), got {}".format(
                    self._base_values.shape[-1], self._window
                ),
                UserWarning,
            )
            return False
        # --- no smoothing when column dimension is not a categorical date ---
        if not self._is_cat_date:
            warnings.warn(
                "No smoothing performed. Column dimension must be a categorical date."
            )
            return False
        return True

    @lazyproperty
    def _is_cat_date(self):
        """True for a categorical dimension having date defined on all valid categories.

        Only meaningful when the dimension is known to be categorical
        (has base-type `categorical`).
        """
        # TODO: change how this categorical date check type when base measure is managed
        # directly from the matrix later on.
        categories = self._partition.smoothed_dimension_dict["type"].get(
            "categories", []
        )
        if not categories:
            return False
        return all(
            category.get("date")
            for category in categories
            if not category.get("missing", False)
        )

    @lazyproperty
    def _smoothed_values(self):
        """1D or 2D float64 ndarray of smoothed base measure.

        In this case the moving average smoother is performed using the np.convolve
        (https://numpy.org/doc/stable/reference/generated/numpy.convolve.html)
        operator that returns the discrete, linear convolution of two one-dimensional
        sequences.
        A moving average is a form of a convolution often used in time series analysis
        to smooth out noise in data by replacing a data point with the average of
        neighboring values in a moving window. A moving average is essentially a
        low-pass filter because it removes short-term fluctuations to highlight a deeper
        underlying trend.
        """
        w = self._window
        values = self._base_values
        return (
            np.array(tuple(np.convolve(values, np.ones(w), mode="valid") / w))
            if values.ndim == 1
            else np.array(
                [tuple(np.convolve(v, np.ones(w), mode="valid") / w) for v in values]
            )
        )

    @lazyproperty
    def _window(self):
        """int, the window parameter specified in the measure expression"""
        window = self._measure_expression.get("window")
        if not window:
            raise ValueError("Window parameter cannot be None.")
        return window



# encoding: utf-8

"""CubeSlice class."""

from __future__ import division

import warnings

import numpy as np
from tabulate import tabulate
from scipy.stats import norm
from scipy.stats.contingency import expected_freq

from cr.cube.enums import DIMENSION_TYPE as DT
from cr.cube.legacy_min_base_size_mask import MinBaseSizeMask
from cr.cube.measures.scale_means import ScaleMeans
from cr.cube.measures.wishart_pairwise_significance import WishartPairwiseSignificance
from cr.cube.measures.old_pairwise_significance import PairwiseSignificance
from cr.cube.util import (
    compress_pruned,
    lazyproperty,
    memoize,
    intersperse_hs_in_std_res,
)

try:
    xrange
except NameError:  # pragma: no cover
    xrange = range


class CubeSlice(object):
    """Two-dimensional projection of a cube.

    The CubeSlice object is used to uniformly represent any cube as one or
    more 2D cubes. For 1D cubes, this is achieved by inflating. For 3D cubes,
    this is achieved by slicing. CubeSlice objects for a cube are accessed
    using the :meth:`.CrunchCube.get_slices` method.
    """

    row_dim_ind = 0

    def __init__(self, cube, index, ca_as_0th=False):

        if ca_as_0th and cube.dim_types[0] != DT.CA_SUBVAR:
            msg = (
                "Cannot set CA as 0th for cube that "
                "does not have CA items as the 0th dimension."
            )
            raise ValueError(msg)

        self._cube = cube
        self._index = index
        self.ca_as_0th = ca_as_0th

    def __repr__(self):
        """Provide text representation suitable for working at console.

        Falls back to a default repr on exception, such as might occur in
        unit tests where object need not otherwise be provided with all
        instance variable values.
        """
        try:
            name = self.table_name if self.table_name else self.name
            dimensionality = " x ".join(dt.name for dt in self.dim_types)
            dim_names = " x ".join(d.name for d in self.dimensions)
            labels = self.labels()
            headers = labels[1] if len(labels) > 1 else ["N"]
            tbl = (
                self.as_array()
                if len(self.as_array().shape) > 1
                else self.as_array()[:, None]
            )
            body = tabulate(
                [[lbl] + lst for lbl, lst in zip(labels[0], tbl.tolist())],
                headers=headers,
            )
            return "%s(name='%s', dim_types='%s', dims='%s')\n%s" % (
                type(self).__name__,
                name,
                dimensionality,
                dim_names,
                body,
            )
        except Exception:
            return super(CubeSlice, self).__repr__()

    def as_array(
        self,
        include_missing=False,
        weighted=True,
        include_transforms_for_dims=None,
        prune=False,
    ):
        hs_dims = self._hs_dims_for_cube(include_transforms_for_dims)
        array = self._cube.as_array(
            weighted=weighted,
            include_missing=include_missing,
            include_transforms_for_dims=hs_dims,
            prune=prune,
        )

        return self._extract_slice_result_from_cube(array)

    def proportions(
        self,
        axis=None,
        weighted=True,
        include_transforms_for_dims=None,
        include_mr_cat=False,
        prune=False,
    ):
        axis = self._calculate_correct_axis_for_cube(axis)
        hs_dims = self._hs_dims_for_cube(include_transforms_for_dims)

        proportions = self._cube.proportions(
            axis=axis,
            weighted=weighted,
            include_mr_cat=include_mr_cat,
            include_transforms_for_dims=hs_dims,
            prune=prune,
        )

        return self._extract_slice_result_from_cube(proportions)

    @lazyproperty
    def dimensions(self):
        return self._cube.dimensions[-2:]

    def inserted_hs_indices(self, prune=False):
        inserted_hs_indices = self._cube.inserted_hs_indices(prune)
        return inserted_hs_indices if self.ca_as_0th else inserted_hs_indices[-2:]

    @lazyproperty
    def is_weighted(self):
        return self._cube.is_weighted

    @lazyproperty
    def name(self):
        if not self.dimensions:
            return None
        return self.dimensions[0].name

    def _prune_indices(self, transforms=None):
        hs_dims = self._hs_dims_for_cube(transforms)
        prune_indices = self._cube.prune_indices(hs_dims)
        return self._extract_slice_result_from_cube(prune_indices)

    @lazyproperty
    def has_means(self):
        return self._cube.has_means

    @lazyproperty
    def is_univariate_ca(self):
        return self._cube.is_univariate_ca

    @lazyproperty
    def population_fraction(self):
        return self._cube.population_fraction

    @lazyproperty
    def univariate_ca_main_axis(self):
        return self._cube.univariate_ca_main_axis

    @lazyproperty
    def ca_main_axis(self):
        """For univariate CA, the main axis is the categorical axis"""
        try:
            ca_ind = self.dim_types.index(DT.CA_SUBVAR)
            return 1 - ca_ind
        except ValueError:
            return None

    @lazyproperty
    def col_dim_ind(self):
        """Return 1 if not categorical array as 0th, 0 otherwise."""
        return 1 if not self.ca_as_0th else 0

    @lazyproperty
    def can_compare_pairwise(self):
        """Return bool indicating if slice can compute pairwise comparisons.

        Currently, only the CAT x CAT slice can compute pairwise comparisons. This also
        includes the categorical array categories dimnension (CA_CAT).
        """
        if self.ndim != 2:
            return False

        return all(dt in DT.ALLOWED_PAIRWISE_TYPES for dt in self.dim_types)

    @lazyproperty
    def dim_types(self):
        """Tuple of DIMENSION_TYPE member for each dimension of slice."""
        return self._cube.dim_types[-2:]

    @memoize
    def get_shape(self, prune=False, hs_dims=None):
        """Tuple of array dimensions' lengths.

        It returns a tuple of ints, each representing the length of a cube
        dimension, in the order those dimensions appear in the cube.
        Pruning is supported. Dimensions that get reduced to a single element
        (e.g. due to pruning) are removed from the returning shape, thus
        allowing for the differentiation between true 2D cubes (over which
        statistical testing can be performed) and essentially
        1D cubes (over which it can't).

        Usage:

        >>> shape = get_shape()
        >>> pruned_shape = get_shape(prune=True)
        """
        if not prune:
            return self.as_array(include_transforms_for_dims=hs_dims).shape

        shape = compress_pruned(
            self.as_array(prune=True, include_transforms_for_dims=hs_dims)
        ).shape
        # Eliminate dimensions that get reduced to 1
        # (e.g. single element categoricals)
        return tuple(n for n in shape if n > 1)

    @lazyproperty
    def has_ca(self):
        """Check if the cube slice has the CA dimension.

        This is used to distinguish between slices that are considered 'normal'
        (like CAT x CAT), that might be a part of the 3D cube that has 0th dim
        as the CA items (subvars). In such a case, we still need to process
        the slices 'normally', and not address the CA items constraints.
        """
        return DT.CA_SUBVAR in self.dim_types

    @lazyproperty
    def has_mr(self):
        """True if the slice has MR dimension.

        This property needs to be overridden, because we don't care about the
        0th dimension (and if it's an MR) in the case of a 3D cube.
        """
        return DT.MR in self.dim_types

    def index_table(self, axis=None, baseline=None, prune=False):
        """Return index percentages for a given axis and baseline.

        The index values represent the difference of the percentages to the
        corresponding baseline values. The baseline values are the univariate
        percentages of the corresponding variable.
        """
        proportions = self.proportions(axis=axis)
        baseline = (
            baseline if baseline is not None else self._prepare_index_baseline(axis)
        )

        # Fix the shape to enable correct broadcasting
        if (
            axis == 0
            and len(baseline.shape) <= 1
            and self.ndim == len(self.get_shape())
            and not self.dim_types == (DT.MR, DT.MR)
        ):
            baseline = baseline[:, None]

        indexes = proportions / baseline * 100

        return self._apply_pruning_mask(indexes) if prune else indexes

    @lazyproperty
    def is_double_mr(self):
        """This has to be overridden from cr.cube.

        If the underlying cube is 3D, the 0th dimension must not be taken into
        account, since it's only the tabs dimension, and mustn't affect the
        properties of the slices.
        """
        return self.dim_types == (DT.MR, DT.MR)

    def labels(self, hs_dims=None, prune=False):
        """Get labels for the cube slice, and perform pruning by slice."""
        if self.ca_as_0th:
            labels = self._cube.labels(include_transforms_for_dims=hs_dims)[1:]
        else:
            labels = self._cube.labels(include_transforms_for_dims=hs_dims)[-2:]

        if not prune:
            return labels

        def prune_dimension_labels(labels, prune_indices):
            """Get pruned labels for single dimension, besed on prune inds."""
            labels = [label for label, prune in zip(labels, prune_indices) if not prune]
            return labels

        labels = [
            prune_dimension_labels(dim_labels, dim_prune_inds)
            for dim_labels, dim_prune_inds in zip(labels, self._prune_indices(hs_dims))
        ]
        return labels

    def margin(
        self,
        axis=None,
        weighted=True,
        include_missing=False,
        include_transforms_for_dims=None,
        prune=False,
        include_mr_cat=False,
    ):
        """Return ndarray representing slice margin across selected axis.

        A margin (or basis) can be calculated for a contingency table, provided
        that the dimensions of the desired directions are marginable. The
        dimensions are marginable if they represent mutualy exclusive data,
        such as true categorical data. For array types the items dimensions are
        not marginable. Requesting a margin across these dimensions
        (e.g. slice.margin(axis=0) for a categorical array cube slice) will
        produce an error. For multiple response slices, the implicit convention
        is that the provided direction scales to the selections dimension of the
        slice. These cases produce meaningful data, but of a slightly different
        shape (e.g. slice.margin(0) for a MR x CAT slice will produce 2D ndarray
        (variable dimensions are never collapsed!)).

        :param axis: Axis across which to sum. Can be 0 (columns margin),
            1 (rows margin) and None (table margin). If requested across
            variables dimension (e.g. requesting 0 margin for CA array) it will
            produce an error.
        :param weighted: Weighted or unweighted counts.
        :param include_missing: Include missing categories or not.
        :param include_transforms_for_dims: Indices of dimensions for which to
               include transformations
        :param prune: Perform pruning based on unweighted counts.
        :returns: (weighed or unweighted counts) summed across provided axis.
            For multiple response types, items dimensions are not collapsed.
        """

        axis = self._calculate_correct_axis_for_cube(axis)
        hs_dims = self._hs_dims_for_cube(include_transforms_for_dims)

        margin = self._cube.margin(
            axis=axis,
            weighted=weighted,
            include_missing=include_missing,
            include_transforms_for_dims=hs_dims,
            prune=prune,
            include_mr_cat=include_mr_cat,
        )

        return self._extract_slice_result_from_cube(margin)

    def min_base_size_mask(self, size, hs_dims=None, prune=False):
        """Returns MinBaseSizeMask object with correct row, col and table masks.

        The returned object stores the necessary information about the base size, as
        well as about the base values. It can create corresponding masks in teh row,
        column, and table directions, based on the corresponding base values
        (the values of the unweighted margins).

        Usage:
        >>> cube_slice = CrunchCube(response).slices[0]  # obtain a valid cube slice
        >>> cube_slice.min_base_size_mask(30).row_mask
        >>> cube_slice.min_base_size_mask(50).column_mask
        >>> cube_slice.min_base_size_mask(22).table_mask
        """
        return MinBaseSizeMask(self, size, hs_dims=hs_dims, prune=prune)

    @lazyproperty
    def mr_dim_ind(self):
        """Get the correct index of the MR dimension in the cube slice."""
        mr_dim_ind = self._cube.mr_dim_ind
        if self._cube.ndim == 3:
            if isinstance(mr_dim_ind, int):
                if mr_dim_ind == 0:
                    # If only the 0th dimension of a 3D is an MR, the sliced
                    # don't actuall have the MR... Thus return None.
                    return None
                return mr_dim_ind - 1
            elif isinstance(mr_dim_ind, tuple):
                # If MR dimension index is a tuple, that means that the cube
                # (only a 3D one if it reached this path) has 2 MR dimensions.
                # If any of those is 0 ind dimension we don't need to include
                # in the slice dimension (because the slice doesn't see the tab
                # that it's on). If it's 1st and 2nd dimension, then subtract 1
                # from those, and present them as 0th and 1st dimension of the
                # slice. This can happend e.g. in a CAT x MR x MR cube (which
                # renders MR x MR slices).
                mr_dim_ind = tuple(i - 1 for i in mr_dim_ind if i)
                return mr_dim_ind if len(mr_dim_ind) > 1 else mr_dim_ind[0]

        return mr_dim_ind

    @lazyproperty
    def ndim(self):
        """Number of slice dimensions

        Returns 2 if the origin cube has 3 or 2 dimensions.  Returns 1 if the
        cube (and the slice) has 1 dimension.  Returns 0 if the cube doesn't
        have any dimensions.
        """
        return min(self._cube.ndim, 2)

    def scale_means(self, hs_dims=None, prune=False):
        """Return list of column and row scaled means for this slice.

        If a row/col doesn't have numerical values, return None for the
        corresponding dimension. If a slice only has 1D, return only the column
        scaled mean (as numpy array). If both row and col scaled means are
        present, return them as two numpy arrays inside of a list.
        """
        scale_means = self._cube.scale_means(hs_dims, prune)

        if self.ca_as_0th:
            # If slice is used as 0th CA, then we need to observe the 1st dimension,
            # because the 0th dimension is CA items, which is only used for slicing
            # (and thus doesn't have numerical values, and also doesn't constitute any
            # dimension of the actual crosstabs that will be created in this case).
            scale_means = scale_means[0][-1]
            if scale_means is None:
                return [None]
            return [scale_means[self._index]]

        return scale_means[self._index]

    @memoize
    def scale_means_margin(self, axis):
        """Get scale means margin for 2D slice.

        This value represents the scale mean of a single variable that
        constitutes a 2D slice. There's one for each axis, if there are
        numerical values on the corresponding (opposite) dimension. The
        numerical values are filtered by non-missing criterium of the
        opposite dimension.
        """
        return ScaleMeans(self).margin(axis)

    @lazyproperty
    def shape(self):
        """Tuple of array dimensions' lengths.

        It returns a tuple of ints, each representing the length of a cube
        dimension, in the order those dimensions appear in the cube.

        This property is deprecated, use 'get_shape' instead. Pruning is not
        supported (supported in 'get_shape').
        """
        deprecation_msg = "Deprecated. Use `get_shape` instead."
        warnings.warn(deprecation_msg, DeprecationWarning)
        return self.get_shape()

    @lazyproperty
    def table_name(self):
        """Get slice name.

        In case of 2D return cube name. In case of 3D, return the combination
        of the cube name with the label of the corresponding slice
        (nth label of the 0th dimension).
        """
        if self._cube.ndim < 3 and not self.ca_as_0th:
            return None

        title = self._cube.name
        table_name = self._cube.labels()[0][self._index]
        return "%s: %s" % (title, table_name)

    def wishart_pairwise_pvals(self, axis=0):
        """Return square symmetric matrix of pairwise column-comparison p-values.

        Square, symmetric matrix along *axis* of pairwise p-values for the
        null hypothesis that col[i] = col[j] for each pair of columns.

        *axis* (int): axis along which to perform comparison. Only columns (0)
        are implemented currently.
        """
        if axis != 0:
            raise NotImplementedError("Pairwise comparison only implemented for colums")
        return WishartPairwiseSignificance.pvals(self, axis=axis)

    def wishart_pairwise_indices(self, axis=0):
        if axis != 0:
            raise NotImplementedError("Pairwise comparison only implemented for colums")
        return WishartPairwiseSignificance.pairwise_indices(self, axis=axis)

    def population_counts(
        self,
        population_size,
        weighted=True,
        include_missing=False,
        include_transforms_for_dims=None,
        prune=False,
    ):
        axis = None if not self.ca_as_0th else 1
        return (
            self.proportions(
                axis=axis,
                weighted=weighted,
                include_mr_cat=include_missing,
                include_transforms_for_dims=include_transforms_for_dims,
                prune=prune,
            )
            * population_size
            * self._cube.population_fraction
        )

    def pvals(self, weighted=True, prune=False, hs_dims=None):
        """Return 2D ndarray with calculated P values

        This function calculates statistically significant cells for
        categorical contingency tables under the null hypothesis that the
        row and column variables are independent (uncorrelated).
        The values are calculated for 2D tables only.

        :param weighted: Use weighted counts for zscores
        :param prune: Prune based on unweighted counts
        :param hs_dims: Include headers and subtotals (as NaN values)
        :returns: 2 or 3 Dimensional ndarray, representing the p-values for each
                  cell of the table-like representation of the crunch cube.
        """
        stats = self.zscore(weighted=weighted, prune=prune, hs_dims=hs_dims)
        pvals = 2 * (1 - norm.cdf(np.abs(stats)))

        return self._apply_pruning_mask(pvals, hs_dims) if prune else pvals

    def zscore(self, weighted=True, prune=False, hs_dims=None):
        """Return ndarray with slices's standardized residuals (Z-scores).

        (Only applicable to a 2D contingency tables.) The Z-score or
        standardized residual is the difference between observed and expected
        cell counts if row and column variables were independent divided
        by the residual cell variance. They are assumed to come from a N(0,1)
        or standard Normal distribution, and can show which cells deviate from
        the null hypothesis that the row and column variables are uncorrelated.

        See also *pairwise_chisq*, *pairwise_pvals* for a pairwise column-
        or row-based test of statistical significance.

        :param weighted: Use weighted counts for zscores
        :param prune: Prune based on unweighted counts
        :param hs_dims: Include headers and subtotals (as NaN values)
        :returns zscore: ndarray representing cell standardized residuals (Z)
        """
        counts = self.as_array(weighted=weighted)
        total = self.margin(weighted=weighted)
        colsum = self.margin(axis=0, weighted=weighted)
        rowsum = self.margin(axis=1, weighted=weighted)
        zscore = self._calculate_std_res(counts, total, colsum, rowsum)

        if hs_dims:
            zscore = intersperse_hs_in_std_res(self, hs_dims, zscore)

        if prune:
            return self._apply_pruning_mask(zscore, hs_dims)

        return zscore

    def pairwise_indices(self, alpha=0.05, only_larger=True, hs_dims=None):
        """Indices of columns where p < alpha for column-comparison t-tests

        Returns an array of tuples of columns that are significant at p<alpha,
        from a series of pairwise t-tests.

        Argument both_pairs returns indices striclty on the test statistic. If
        False, however, only the index of values *significantly smaller* than
        each cell are indicated.
        """
        return PairwiseSignificance(
            self, alpha=alpha, only_larger=only_larger, hs_dims=hs_dims
        ).pairwise_indices

    def summary_pairwise_indices(self, alpha=0.05, only_larger=True, hs_dims=None):
        return PairwiseSignificance(
            self, alpha=alpha, only_larger=only_larger, hs_dims=hs_dims
        ).summary_pairwise_indices

    def pairwise_significance_tests(self, column_idx, hs_dims=None):
        """list of _ColumnPairwiseSignificance tests.

        Result has as many elements as there are columns in the slice. Each
        significance test contains `p_vals` and `t_stats` (ndarrays that represent
        probability values and statistical scores).
        """
        return PairwiseSignificance(self, hs_dims=hs_dims).values[column_idx]

    def _apply_pruning_mask(self, res, hs_dims=None):
        array = self.as_array(prune=True, include_transforms_for_dims=hs_dims)

        if not isinstance(array, np.ma.core.MaskedArray):
            return res

        return np.ma.masked_array(res, mask=array.mask)

    def _array_type_std_res(self, counts, total, colsum, rowsum):
        """Return ndarray containing standard residuals for array values.

        The shape of the return value is the same as that of *counts*.
        Array variables require special processing because of the
        underlying math. Essentially, it boils down to the fact that the
        variable dimensions are mutually independent, and standard residuals
        are calculated for each of them separately, and then stacked together
        in the resulting array.
        """
        if self.mr_dim_ind == 0:
            # --This is a special case where broadcasting cannot be
            # --automatically done. We need to "inflate" the single dimensional
            # --ndarrays, to be able to treat them as "columns" (essentially a
            # --Nx1 ndarray). This is needed for subsequent multiplication
            # --that needs to happen column wise (rowsum * colsum) / total.
            total = total[:, np.newaxis]
            rowsum = rowsum[:, np.newaxis]

        expected_counts = rowsum * colsum / total
        variance = rowsum * colsum * (total - rowsum) * (total - colsum) / total ** 3
        return (counts - expected_counts) / np.sqrt(variance)

    def _calculate_std_res(self, counts, total, colsum, rowsum):
        """Return ndarray containing standard residuals.

        The shape of the return value is the same as that of *counts*.
        """
        if set(self.dim_types) & DT.ARRAY_TYPES:  # ---has-mr-or-ca---
            return self._array_type_std_res(counts, total, colsum, rowsum)
        return self._scalar_type_std_res(counts, total, colsum, rowsum)

    def _calculate_correct_axis_for_cube(self, axis):
        """Return correct axis for cube, based on ndim.

        If cube has 3 dimensions, increase axis by 1. This will translate the
        default 0 (cols direction) and 1 (rows direction) to actual 1
        (cols direction) and 2 (rows direction). This is needed because the
        0th dimension of the 3D cube is only used to slice across. The actual
        margins need to be calculated for each slice separately, and since
        they're implemented as an ndarray, the direction needs to be increased
        by one. For the value of `None`, don't modify the axis parameter.

        :param axis: 0, 1, or None. Axis that will be passed to self._cube
                     methods. If the cube is 3D, the axis is typically
                     increased by 1, to represent correct measure direction.
        :returns: int or None, representing the updated axis to pass to cube
        """

        if self._cube.ndim < 3:
            if self.ca_as_0th and axis is None:
                # Special case for CA slices (in multitables). In this case,
                # we need to calculate a measurement across CA categories
                # dimension (and not across items, because it's not
                # allowed). The value for the axis parameter of None, would
                # imply both cat and items dimensions, and we don't want that.
                return 1
            return axis

        # Expected usage of the 'axis' parameter from CubeSlice is 0, 1, or
        # None. CrunchCube handles all other logic. The only 'smart' thing
        # about the handling here, is that the axes are increased for 3D cubes.
        # This way the 3Dness is hidden from the user and he still sees 2D
        # crosstabs, with col and row axes (0 and 1), which are transformed to
        # corresponding numbers in case of 3D cubes (namely 1 and 2). In the
        # case of None, we need to analyze across all valid dimensions, and the
        # CrunchCube takes care of that (no need to update axis if it's None).
        # If the user provides a tuple, it's considered that he "knows" what
        # he's doing, and the axis argument is not updated in this case.
        if isinstance(axis, int):
            axis += 1
        return axis

    def _hs_dims_for_cube(self, hs_dims):
        if self._cube.ndim < 3:
            return hs_dims

        # Keep the 2D illusion for the user. If a user sees a 2D slice, he
        # still needs to be able to address both dimensions (for which he
        # wants the H&S included) as 0 and 1. Since these are offset by a 0
        # dimension in a 3D case, inside the cr.cube, we need to increase
        # the indexes of the required dims.
        return [d + 1 for d in hs_dims] if hs_dims is not None else None

    def _prepare_index_baseline(self, axis):
        # First get the margin of the opposite direction of the index axis.
        # We need this in order to end up with the right shape of the
        # numerator vs denominator.
        baseline = self.margin(axis=(1 - axis), include_missing=True)
        if len(baseline.shape) <= 1:
            # If any dimension gets flattened out, due to having a single
            # element, re-inflate it
            baseline = baseline[None, :]
        slice_ = [slice(None)]
        total_axis = None
        if isinstance(self.mr_dim_ind, tuple):
            if self.get_shape()[0] == 1 and axis == 0:
                total_axis = axis
                slice_ = [0]
            elif self.get_shape()[0] == 1 and axis == 1:
                total_axis = 1
                slice_ = [slice(None), 0]
            else:
                total_axis = axis + 1
                slice_ += [slice(None), 0] if axis == 1 else [0]
        elif self.mr_dim_ind is not None:
            slice_ = [0] if self.mr_dim_ind == 0 and axis != 0 else [slice(None), 0]
            total_axis = axis if self.mr_dim_ind != 0 else 1 - axis

        total = np.sum(baseline, axis=total_axis)
        baseline = baseline[tuple(slice_)]

        if axis == self.mr_dim_ind:
            return baseline / total
        elif isinstance(self.mr_dim_ind, tuple) and axis in self.mr_dim_ind:
            return baseline / total

        baseline = baseline if len(baseline.shape) <= 1 else baseline[0]
        baseline = baseline / np.sum(baseline)
        return baseline / np.sum(baseline, axis=0)

    def _scalar_type_std_res(self, counts, total, colsum, rowsum):
        """Return ndarray containing standard residuals for category values.

        The shape of the return value is the same as that of *counts*.
        """
        expected_counts = expected_freq(counts)
        residuals = counts - expected_counts
        variance = (
            np.outer(rowsum, colsum)
            * np.outer(total - rowsum, total - colsum)
            / total ** 3
        )
        return residuals / np.sqrt(variance)

    def _extract_slice_result_from_cube(self, result):
        if self._cube.ndim < 3 and not self.ca_as_0th or len(result) - 1 < self._index:
            return result
        result = result[self._index]
        if isinstance(result, tuple):
            return np.array(result, dtype=object)
        elif not isinstance(result, np.ndarray):
            result = np.array([result])
        return result



# encoding: utf-8

"""The cube data-partion object used for 0D means cube."""

from __future__ import absolute_import, division, print_function, unicode_literals

from cr.cube.util import lazyproperty


class MeansScalar(object):
    """Represents slices with means (and no counts)."""

    def __init__(self, means, unweighted_counts):
        self._means = means
        self._unweighted_counts = unweighted_counts

    @lazyproperty
    def means(self):
        return self._means

    @lazyproperty
    def ndim(self):
        """int count of dimensions in this scalar, unconditionally 0.

        A scalar is by definition zero-dimensional.
        """
        return 0

    @lazyproperty
    def table_base(self):
        # TODO: Check why we expect mean instead of the real base in this case.
        return self.means
